import numpy as np

def one_hot(class_idx: int, total_classes: int) -> np.ndarray:
    '''
        Performs one hot enoding on the index.
        
        :param class_idx: index of the selected class
        :param total_classes: total number of classes
        :returns: 2D aray of one hot encoded indices.
    '''
    
    index = np.asarray(class_idx)
    
    if len(index.shape) == 0:
        index = np.asarray([index])
    
    assert len(index.shape) == 1
    
    num = index.shape[0]
    output = np.zeros((num, total_classes), dtype=np.float32)
    output[np.arange(num), index] = 1
    
    return output

def one_hot_if_needed(class_idx, total_classes) -> np.ndarray:
    '''
        Checks if the labels are already one hot encoded.
        Performs One hot encoding only if the labels are not already onehot encoded.

        :param class_idx: index of the selected class
        :param total_classes: total number of classes
        :returns: 2D aray of one hot encoded indices.
    '''
    label = np.asarray(class_idx)
    
    if len(label.shape) <= 1:
        label = one_hot(class_idx, total_classes)
    
    assert len(label.shape) == 2
    return label

## GAN Utils

from scipy.stats import truncnorm


def noise_sample(n_samples, dim, limits = (-1, 1)):
    '''
    Generates noise vector of given dimension.

    :param n_samples: number of noise vector samples to be generated
    :param dim: dimension of the noise vector to be generated
    :param limits: lower and upper limits of the values in the noise vector
    :return: noise sample of given dimension
    '''
    noise = np.random.uniform(limits[0], limits[1], size=(n_samples, dim))
    return noise


def interpolate(A, B, num_interps):
    '''
        Find interpolations between two given vectors.

        :param A: 1st vector
        :param B: 2nd vector
        :param num_interps: number of interpolations to make
        :return: an array of interpolated noise vectors
    '''
    
    if A.shape != B.shape:
        raise ValueError('A and B must have the same shape to interpolate.')
    alphas = np.linspace(0, 1, num_interps)
    return np.array([(1-a)*A + a*B for a in alphas])

def interpolate_and_shape(A, B, num_samples, num_interps):
    interps = interpolate(A, B, num_interps)
    return (interps.transpose(1, 0, *range(2, len(interps.shape))).reshape(num_samples * num_interps, *interps.shape[2:]))


def truncated_noise_sample(n_samples, dim, truncation=1., seed=None):
    '''
    Generates truncated noise samples.
    
    :param n_samples: number of noise vector samples to be generated
    :param dim: dimension of the noise vector to be generated
    :param truncation: truncation value for the noise vector
    :param seed: random seed to use
    :return: noise sample of given dimension and truncation
    '''

    state = None if seed is None else np.random.RandomState(seed)
    values = truncnorm.rvs(-2, 2, size=(n_samples, dim), random_state=state)
    return truncation * values
    
# Experimental GAN utils


from typing import Optional, Union, List
import numpy as np

import tensorflow as tf
import tensorflow_hub as hub

from deepdraken.utils import one_hot, one_hot_if_needed, truncated_noise_sample, interpolate_and_shape


class BigGAN():

    '''
        Class for generating big gan images.
    '''
    MODELS = { 'biggan-deep-128': 'https://tfhub.dev/deepmind/biggan-deep-128/1',  # 128x128 BigGAN-deep
               'biggan-deep-256': 'https://tfhub.dev/deepmind/biggan-deep-256/1',  # 256x256 BigGAN-deep
               'biggan-deep-512': 'https://tfhub.dev/deepmind/biggan-deep-512/1',  # 512x512 BigGAN-deep
               'biggan-128': 'https://tfhub.dev/deepmind/biggan-128/2',  # 128x128 BigGAN
               'biggan-256': 'https://tfhub.dev/deepmind/biggan-256/2',  # 256x256 BigGAN
               'biggan-512': 'https://tfhub.dev/deepmind/biggan-512/2'}  # 512x512 BigGAN

    def __init__(self, model_name: Optional[str] = 'biggan-deep-512') -> None:

        self.model = hub.KerasLayer(self.MODELS[model_name])

    def __get_images(self, y: np.ndarray, z: np.ndarray, truncation: int) -> np.ndarray:
        '''
        Method for generating images.
        :param y: one hot encoded class labels
        :param z: noise vectors
        :param truncations: truncation value for the noise vector
        :return: generated images
        '''
        # generating and post processing the image
        images = self.model({'y': y, 'z':z, 'truncation': truncation})
        images = np.asarray(images)
        images = np.clip(((images + 1) / 2.0) * 256, 0, 255)
        images = np.uint8(images)
        return images

    def sample(self,
               label: List[int],
               truncation: Optional[float] = 0.4,
               noise_seed: Optional[int] = None):
        '''
        Method for sampling images

        :param label: class of the image to be generated
        :param num_samples: number of images to generate
        :param truncation: truncation value for the noise vector
        :param noise_seed: seed for generating the noise, random seed is taken if None
        :param plot: plots the images if set to true
        :return: an array of generated images
        '''
        num_samples = len(label)
    
        # generating the noise
        z = truncated_noise_sample(num_samples, 128, truncation, noise_seed)
        # converting into array if needed
        z = np.asarray(z)
        label = np.asarray(label)
        
        y = one_hot_if_needed(label, 1000) # one hot encoding the label
        
        images = self.__get_images(y, z, truncation)

        return images

    def interpolate(self,
                    label_A: str,
                    label_B: str,
                    num_samples: int,
                    num_interps: Optional[int] = 5,
                    truncation: Optional[float] = 0.2,
                    noise_seed_A: Optional[int] = None,
                    noise_seed_B: Optional[int] = None):
        '''
        Generates interpolations for images between two classes or labels.
        
        :param label_A: class 1 for interpolation
        :param label_B: class 2 for interpolation
        :param num_samples: number of samples to generate
        :param num_interps: number of interpolations to generate between the classes
        :param truncation: truncation value for the noise vector
        :param noise_seed_A: seed for generating noise for the 1st class
        :param noise_seed_B: seed for generating noise for the 2nd class
        :return: an array of array containing interpolated images
        '''

        # generating noise samples of shape num_samples, 128 each
        z_A, z_B = [truncated_noise_sample(num_samples, 128, truncation, noise_seed) for noise_seed in [noise_seed_A, noise_seed_B]]
        # generating one_hot encoded class vectors of the class
        y_A, y_B = [one_hot([category] * num_samples, 1000) for category in [label_A, label_B]]

        # interpolating the noise samples and class vectors
        z_interp = interpolate_and_shape(z_A, z_B, num_samples, num_interps)
        y_interp = interpolate_and_shape(y_A, y_B, num_samples, num_interps)

        # generating and reshaping the image
        images = self.__get_images(y_interp, z_interp, truncation)
        shape = [num_samples, num_interps]
        shape.extend([i for i in images.shape[1:]])
        images = images.reshape(shape)

        return images


from typing import Union
import numpy as np
from statistics import mean
import itertools

import torch
from torch import nn
import torchvision as tv

from deepdraken.base import BaseModel

# TODO class for dataloader

class DCGAN(BaseModel):

    def __init__(self,
                 net_G,
                 net_D,
                 device='cpu',
                 gpu_ids=None):
        
        super().__init__(device, gpu_ids)

        self.net_G = net_G
        self.net_D = net_D

        self.net_G.to(self.device)
        self.net_D.to(self.device)
        
        if 'cuda' in str(device):
            self.net_G = nn.DataParallel(self.net_G, self.gpu_ids)
            self.net_D = nn.DataParallel(self.net_D, self.gpu_ids)
        
        self.z_dim = next(self.net_G.parameters()).shape[-1]
        
        with torch.no_grad():
            self.c, self.w, self.h = self.net_G(self.get_noise(1, self.z_dim)).shape[1:]
        
        print(f'Detected Data:',
                f'\nNoise dimention    : {self.z_dim}', 
                f'\nNumber of Channels : {self.c}',
                f'\nImage Height       : {self.h}',
                f'\nImage Width        : {self.w}')

        self.model_names = ['net_G', 'net_D']
        self.optimizer_names = ['optimizer_G', 'optimizer_D']
        self.loss_names = ['loss_G', 'loss_D', 'loss_D_real', 'loss_D_fake']
        self.metric_names = []
        self.optimizers = []
        
        self.loss_G = float('nan')
        self.loss_D, self.loss_D_real, self.loss_D_fake = float('nan'), float('nan'), float('nan')

        self.history = {}
        self.batch_history = {}
        for loss_name in self.loss_names:
            self.history[loss_name] = []
            self.batch_history[loss_name] = []

    def set_data(self, batch) -> None:

        '''
        Unpacks a single batch data from the dataloader and perform necessary steps.
        Takes in a single batch as input.
        :param batch: single batch of data
        :return: None
        '''
        
        real_images, labels = batch
        self.real_images = real_images.to(self.device)
        self.labels = labels.to(self.device)
        self.batch_size = real_images.shape[0]
        
        self.valid = torch.ones((self.batch_size, 1)).to(self.device)
        self.fake = torch.zeros_like(self.valid)
        self.noise = self.get_noise(self.batch_size, self.z_dim)

    def get_noise(self, batch_size: int, z_dim: int) -> torch.Tensor:
        return torch.randn(batch_size, z_dim).to(self.device)

    def get_dataloader(self, root, batch_size, shuffle = True, transform = None):

        def get_default_transform():
            transform = []

            if self.c == 1:
                transform.append(tv.transforms.Grayscale(num_output_channels=1))

            transform.extend([tv.transforms.Resize((self.h, self.w)),
                              tv.transforms.ToTensor(),
                              tv.transforms.Normalize((0.5,) * self.c, (0.5,) * self.c),])
            return tv.transforms.Compose(transform)

        if transform == None :
            transform = get_default_transform()
        
        dataset = tv.datasets.ImageFolder(root=root, transform=transform)
        print(f'Class to index mapping:\n{dataset.class_to_idx}')
        dataloader = torch.utils.data.DataLoader(dataset, batch_size=batch_size, shuffle=shuffle)
        return dataloader

    def forward(self) -> None:
        '''
        Executes forward pass; called by both functions <optimize_parameters> and <test>.
        :return: None
        '''
        
        self.fake_images = self.net_G(self.noise)

    def backward_G(self) -> None:
        '''
        Executes backward pass for generator.
        '''
        
        self.loss_G = self.criterion(self.net_D(self.fake_images), self.valid)
        self.loss_G.backward()

    def backward_D(self) -> None:
        '''
        Executes backward pass for discriminator.
        '''
        
        self.loss_D_real = self.criterion(self.net_D(self.real_images), self.valid)
        self.loss_D_fake = self.criterion(self.net_D(self.fake_images.detach()), self.fake)
        self.loss_D = (self.loss_D_real + self.loss_D_fake) / 2
        self.loss_D.backward()

    def optimize_parameters(self) -> None:
        '''
        Calculate losses, gradients, and update network weights; called in every training iteration
        :return: None
        '''
        
        self.forward()      # compute fake images

        # -----------------
        #  Train Generator
        # -----------------
        self.set_requires_grad([self.net_D], False)     # D require no gradients when optimizing G
        self.optimizer_G.zero_grad()    # set G's gradients to zero
        self.backward_G()               # calculate gradients for G
        self.optimizer_G.step()         # update G's weights

        # ---------------------
        #  Train Discriminator
        # ---------------------
        self.set_requires_grad([self.net_D], True)
        self.optimizer_D.zero_grad()    # set D's gradients to zero
        self.backward_D()               # calculate gradients for D
        self.optimizer_D.step()         # update D's weights

    def compile(self,
                optim_G, 
                optim_D,
                criterion = nn.BCELoss()) -> None:
        '''
        Method for setting the optimizers, lr schedulers, etc.
        
        :param gen_optim: generator optimizer class
        :param disc_optim: discriminator optimizer class
        :param criterion: loss criterion, default: Binary Crossentropy
        :return: None
        '''

        self.optimizer_G = optim_G
        self.optimizer_D = optim_D
        self.optimizers = [self.optimizer_G, self.optimizer_D]

        self.criterion = criterion.to(self.device)

    def fit(self, 
            data : Union[str, torch.Tensor], 
            epochs: int, 
            batch_size: int = 64, 
            shuffle: bool = True, 
            transform: tv.transforms.Compose = None,) -> None:
        '''
        Trains the models
        :param data: dataloader | path to dataset | pytorch tensor
        :param epochs (int): number of epochs to train the model
        :param batch_size (int): batch size to use while training, default: 64
        :param shuffle (bool): to shuffle the dataset or not, default: True
        :param transform : transformations to be applied on the dataset, i.e. preprocessing, default: None

        Note: batch_size , shuffle and transforms are applicable only when path to datset or tensor is proved as data
              in case dataloader is provided these parameters are handeled by dataloader itself.

        :return: None
        '''

        # Setting nets to training mode
        self.set_mode('train')

        if isinstance(data, str):
            print(f'String input detected.\nCreating a Dataloader using {data} directory')
            data = self.get_dataloader(data, batch_size, shuffle, transform)
        elif isinstance(data, torch.Tensor):
            # TODO for direct tensor inputs
            raise NotImplementedError()

        for epoch in range(epochs): # TODO use TQDM here
            self.on_epoch_start()
            
            for batch_idx, batch in enumerate(data):
                self.set_data(batch)
                self.optimize_parameters()
                self.append_batch_history() 

            self.append_history()
            print("[Epoch {}/{}] [G loss: {}] [D loss: {}]".format(epoch, epochs, self.history['loss_G'][-1], self.history['loss_D'][-1]))
            self.on_epoch_end()

    def generate_image(self, num_sample:int = 1, dtype:str ='tensor') -> Union[torch.Tensor, np.ndarray]:
        '''
        Generates images using Generator model.
        :param num_sample (int): number of images to be generated
        :param type (str): datatype of the returned image
                                -- tensor : pytorch tensor
                                -- numpy  : numpy array
        :return (torch.Tensor or numpy.ndarray): generated image pytorh tensor or numpy array
        '''

        noise = self.get_noise(num_sample, self.z_dim)
        with torch.no_grad():
            img = self.net_G(noise)

        img = img.cpu()

        if dtype == "tensor":
            return img
        elif dtype == "array":
            return img.numpy()
        else:
            print(f"Image of {dtype} datatype can not be generaed. Please choose from tensor or array")

    def on_epoch_start(self):
        pass

    def on_epoch_end(self):
        pass

    def append_batch_history(self):
        for name in itertools.chain(self.loss_names, self.metric_names):
            self.batch_history[name].append(float(getattr(self, name)))

    def append_history(self):

        for name in itertools.chain(self.loss_names, self.metric_names):
            self.history[name].append(mean(self.batch_history[name]))
            self.batch_history[name] = []

# TODO functionality of using lr schedulers
# TODO functionality of using data other than from dataloader, e.g. tensors
# TODO interpolating generated images --> linear, spherical
# TODO saving images in disc
# TODO loading only generator model for sampling if training mode not available.
# TODO functionality for custom noise function
# TODO gradient clipping
# TODO gradient penalty



import torch

from deepdraken.image_generation.gans.dcgan import DCGAN

class WGAN(DCGAN):

    def __init__(self, net_G, net_D, device='cpu', gpu_ids=None):
        super().__init__(net_G, net_D, device, gpu_ids)

    def backward_G(self):
        
        self.loss_G = -torch.mean(self.net_D(self.fake_images))
        self.loss_G.backward()

    def backward_D(self):
        
        self.loss_D_real = -torch.mean(self.net_D(self.real_images))
        self.loss_D_fake = torch.mean(self.net_D(self.fake_images.detach()))
        self.loss_D = self.loss_D_real + self.loss_D_fake # TODO check if division by 2 is to be done or not
        self.loss_D.backward()
    
    def optimize_parameters(self) -> None:
         
        data = super().optimize_parameters()
        for p in self.net_D.parameters():
            p.data.clamp_(self.lower_clip, self.upper_clip)

        return data

    def compile(self, optim_G, optim_D):
        self.optimizer_G = optim_G
        self.optimizer_D = optim_D

    def fit(self, data, epochs, batch_size=64, shuffle=True, transform=None, clip_value=0.01):

        if isinstance(clip_value, float):
            self.upper_clip = abs(clip_value)
            self.lower_clip = -self.upper_clip
        if isinstance(clip_value, tuple):
            self.upper_clip, self.lower_clip == clip_value

        return super().fit(data, epochs, batch_size, shuffle, transform)

class WGANGP(WGAN):

    def __init__(self, net_G, net_D, device, is_train=True, gpu_ids=None):
        super().__init__(net_G, net_D, device, is_train, gpu_ids)

    def compute_gradient_penalty(self, real_samples, fake_samples):
        
        """
        Calculates the gradient penalty for WGAN GP
        
        :param real_samples: real images from the dataset
        :param fake_samples: images generated by the generator
        :return: gradient penalty for the model
        """
        
        # Random weight term for interpolation between real and fake samples
        # Random numbers from a uniform distribution on the interval [0, 1)[0,1)
        alpha = torch.rand((self.batch_size, 1, 1, 1), device=self.device)
        
        # Get random interpolation between real and fake samples
        interpolates = (alpha * real_samples + ((1 - alpha) * fake_samples)).requires_grad_(True)
        d_interpolates = self.disc(interpolates)
        fake = torch.ones((self.batch_size, 1)).to(self.device)
        
        # Get gradient w.r.t. interpolates
        gradients = torch.autograd.grad( outputs=d_interpolates,
                                         inputs=interpolates,
                                         grad_outputs=fake,
                                         create_graph=True,
                                         retain_graph=True,
                                         only_inputs=True, )[0]

        gradients = gradients.view(gradients.size(0), -1)
        gradient_penalty = ((gradients.norm(2, dim=1) - 1) ** 2).mean()
        return gradient_penalty

    def backward_D(self):
        
        self.loss_D_real = -torch.mean(self.net_D(self.real_images))
        self.loss_D_fake = torch.mean(self.net_D(self.fake_images.detach()))

        if self.lambda_gp == 0:
            self.loss_D = self.loss_D_real + self.loss_D_fake
        else:
            gradient_penalty = self.compute_gradient_penalty(self.real_images, self.fake_images)
            self.loss_D = self.loss_D_real + self.loss_D_fake + self.lambda_gp * gradient_penalty
        
        self.loss_D.backward()

    def optimize_parameters(self) -> None:
        return super(WGAN, self).optimize_parameters()

    def fit(self, data, epochs, batch_size=64, shuffle=True, transform=None, lambda_gp=10):

        self.lambda_gp = lambda_gp
        return super(WGAN, self).fit(data, epochs, batch_size, shuffle, transform)
        
        

# TODO loss vaue is comming negative, check about it
# TODO loss must be 1L continuous and thus weight clip or gradient penalty is must
# TODO weight cliping is applied only on the critic


from torch.optim import lr_scheduler

def get_scheduler(optimizer, lr_policy, **kwargs):
    """Return a learning rate scheduler
    
    :param optimizer: the optimizer of the network
    :param lr_policy: name of learning rate policy: linear | step | plateau | cosine
    :param kwargs: keyword arguments to set for learning rate
                        linear : epoch_count | n_epochs | n_epochs_decay
                        step   : n_decay_iters
                        plateau: 
                        cosine : n_epochs
    
    For 'linear', we keep the same learning rate for the first <n_epochs> epochs
    and linearly decay the rate to zero over the next <opt.n_epochs_decay> epochs.
    For other schedulers (step, plateau, and cosine), we use the default PyTorch schedulers.
    See https://pytorch.org/docs/stable/optim.html for more details.
    """
    if lr_policy == 'linear':
        def lambda_rule(epoch):
            lr_l = 1.0 - max(0, epoch + kwargs['epoch_count'] - kwargs['n_epochs']) / float(kwargs['n_epochs_decay'] + 1)
            return lr_l
        scheduler = lr_scheduler.LambdaLR(optimizer, lr_lambda=lambda_rule)
    elif lr_policy == 'step':
        scheduler = lr_scheduler.StepLR(optimizer, step_size=kwargs['lr_decay_iters'], gamma=0.1)
    elif lr_policy == 'plateau':
        scheduler = lr_scheduler.ReduceLROnPlateau(optimizer, mode='min', factor=0.2, threshold=0.01, patience=5)
    elif lr_policy == 'cosine':
        scheduler = lr_scheduler.CosineAnnealingLR(optimizer, T_max=kwargs['n_epochs'], eta_min=0)
    else:
        return NotImplementedError('learning rate policy [%s] is not implemented', lr_policy)
    return scheduler

import random
import torch


class ImagePool():
    """This class implements an image buffer that stores previously generated images.

    This buffer enables us to update discriminators using a history of generated images
    rather than the ones produced by the latest generators.
    """

    def __init__(self, pool_size):
        """Initialize the ImagePool class
        
        Parameters:
            pool_size (int) -- the size of image buffer, if pool_size=0, no buffer will be created
        """
        self.pool_size = pool_size
        if self.pool_size > 0:  # create an empty pool
            self.num_imgs = 0
            self.images = []

    def query(self, images):
        """Return an image from the pool.
        
        :param images: the latest generated images from the generator
        
        :return: images from the buffer.
        
        By 50/100, the buffer will return input images.
        By 50/100, the buffer will return images previously stored in the buffer,
        and insert the current images to the buffer.
        """
        if self.pool_size == 0:  # if the buffer size is 0, do nothing
            return images
        return_images = []
        for image in images:
            image = torch.unsqueeze(image.data, 0)
            if self.num_imgs < self.pool_size:   # if the buffer is not full; keep inserting current images to the buffer
                self.num_imgs = self.num_imgs + 1
                self.images.append(image)
                return_images.append(image)
            else:
                p = random.uniform(0, 1)
                if p > 0.5:  # by 50% chance, the buffer will return a previously stored image, and insert the current image into the buffer
                    random_id = random.randint(0, self.pool_size - 1)  # randint is inclusive
                    tmp = self.images[random_id].clone()
                    self.images[random_id] = image
                    return_images.append(tmp)
                else:       # by another 50% chance, the buffer will return the current image
                    return_images.append(image)
        return_images = torch.cat(return_images, 0)   # collect all the images and return
        return return_images


import os
import re
from io import open

from setuptools import find_packages, setup


def parse_requirements(filename):
    """
    Parse a requirements pip file returning the list of required packages. It exclude commented lines and --find-links directives.
    
    :param filename: pip requirements requirements
    :return: list of required package with versions constraints
    """
    with open(filename) as file:
        parsed_requirements = file.read().splitlines()
    parsed_requirements = [line.strip()
                           for line in parsed_requirements
                           if not ((line.strip()[0] == "#") or line.strip().startswith('--find-links') or ("git+https" in line))]
    
    return parsed_requirements


def get_dependency_links(filename):
    """
    Parse a requirements pip file looking for the --find-links directive.
    
    :param filename:  pip requirements requirements
    :return: list of find-links's url
    """
    with open(filename) as file:
        parsed_requirements = file.read().splitlines()
    dependency_links = list()
    for line in parsed_requirements:
        line = line.strip()
        if line.startswith('--find-links'):
            dependency_links.append(line.split('=')[1])
    return dependency_links


dependency_links = get_dependency_links('requirements.txt')
parsed_requirements = parse_requirements('requirements.txt')


def versionfromfile(*filepath):
    infile = os.path.join(*filepath)
    with open(infile) as fp:
        version_match = re.search(
                r"^__version__\s*=\s*['\"]([^'\"]*)['\"]", fp.read(), re.M
        )
        if version_match:
            return version_match.group(1)
        raise RuntimeError("Unable to find version string in {}.".format(infile))


here = os.path.abspath(os.path.dirname(__file__))
_version: str = versionfromfile(here, "deepdraken", "_version.py")

setup(
    name="deepdraken",
    version=_version,
    author="Devavrat Singh Bisht",
    author_email="devavratsinghbisht@gmail.com",
    description="Package for CV tasks",
    long_description=open("README.md", "r", encoding="utf-8").read(),
    long_description_content_type="text/markdown",
    keywords="Computer-Vision Transfer-Learning",
    license="Apache License 2.0",
    url="https://github.com/DevavratSinghBisht/deepdraken",
    download_url=f"https://github.com/DevavratSinghBisht/deepdraken", # TODO see this
    packages=find_packages(exclude=["*.tests", "*.tests.*", "tests.*", "tests"]), # TODO see this
    dependency_links=dependency_links,
    install_requires=parsed_requirements,
    python_requires=">=3.6.0",
    classifiers=[
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: Apache Software License",
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
    ],
)


from typing import List
from pathlib import Path

import torch
import torchvision as tv

from deepdraken.image_generation.gans.dcgan import DCGAN

class CGAN(DCGAN):

    def __init__(self,
                 gen,
                 disc,
                 n_classes:int,
                 device : str = 'cpu',
                 gpu_ids : List[int] = None):
                 
        super().__init__(gen, disc, device, gpu_ids)
        self.n_classes = n_classes

    def set_data(self, batch):
        super().set_data(batch)
        self.fake_labels = torch.randint(0, self.n_classes, (self.batch_size,) , device=self.device)

    def forward(self):
        
        self.fake_images = self.net_G(self.noise, self.labels)

    def backward_G(self):
        
        self.loss_G = self.criterion(self.net_D(self.fake_images, self.fake_labels), self.valid)
        self.loss_G.backward()

    def backward_D(self):
        
        self.loss_D_real = self.criterion(self.net_D(self.real_images, self.labels), self.valid)
        self.loss_D_fake = self.criterion(self.net_D(self.fake_images.detach(), self.fake_labels.detach()), self.fake)
        self.loss_D = (self.loss_D_real + self.loss_D_fake) / 2
        self.loss_D.backward()


from collections import OrderedDict
from abc import ABC, abstractclassmethod, abstractmethod
from typing import List
from pathlib import Path

import torch
from torch import nn
from .utils.base_utils import get_scheduler

class BaseModel(ABC):
    """
    This class is an abstract base class (ABC) for models.
    To create a subclass, you need to implement the following five functions:
        -- <__init__>               :   initialize the class; first call BaseModel.__init__(self, opt).
        -- <set_data>               :   unpack data from dataset and apply preprocessing.
        -- <forward>                :   produce intermediate results.
        -- <optimize_parameters>    :   calculate losses, gradients, and update network weights.
        -- <complie>                :   intializes the optimizers
        -- <fit>                    :   trains the models
    """

    def __init__(self,
                 device : str = 'cpu',
                 gpu_ids: List[int] = None):
        """
        Initialize the BaseModel class.
        
        :param device (str): device (cpu or cuda) to use
        :param gpu_ids (list of int): CUDA devices (default: all devices) 
        
        When creating your custom class, you need to implement your own initialization.
        In this function, you should first call <BaseModel.__init__(self, device, gpu_ids)>
        Then, you need to define four lists:
            -- self.model_names (str list)      :   specify the networks used in training.
            -- self.loss_names (str list)       :   specify the training losses that you want to plot and save.
            -- self.optimizer_names (str list)  :   specifgy the optimizers. 
            -- self.optimizers (optimizer list) :   You might want to keep it as empty list and update the list when self.compile function is called. You can define one optimizer for each network here. If two networks are updated at the same time, you can use itertools.chain to group them. See cycle_gan.py for an example.
        """

        self.gpu_ids = gpu_ids

        # get device name: CPU or GPU
        if device == "cuda":
            if torch.cuda.is_available():
                
                if self.gpu_ids == None:
                    self.gpu_ids = [i for i in range(torch.cuda.device_count())]

                self.device = torch.device(f"cuda:{self.gpu_ids[0]}")
            
            else:
                print("CUDA not available falling back to CPU.")
                self.device = torch.device("cpu")

        elif device == "cpu":
            self.device = torch.device("cpu")
        
        #  initializing lists
        self.model_names = []
        self.loss_names = []
        self.optimizer_names = []
        self.optimizers = []
        # self.visual_names = []
        # self.image_paths = []
        self.metric = 0  # used for learning rate policy 'plateau'
        self.epoch = 0

    @abstractmethod
    def set_data(self, batch: tuple):
        """
        Unpacks a single batch data from the dataloader and perform necessary steps.
        
        :param batch (tuple): single batch of data, (X, y)
        :return: None
        """
        pass

    @abstractmethod
    def forward(self):
        """
        Run forward pass; called by both functions <optimize_parameters> and <test>.
        """
        pass

    @abstractmethod
    def optimize_parameters(self):
        """
        Calculate losses, gradients, and update network weights; called in every training iteration
        """
        pass

    @abstractmethod
    def compile(self,):
        """
        Used to set and initialize optimizers, loss, metrics and lr schedulers.
        """
        pass

    @abstractmethod
    def fit(self,):
        """
        Trains the networks
        """
        pass

    # def set_scheduler(self, lr_policy: str, schedulers = [], **kwargs) -> None:
    #     """
    #     Create schedulers
        
    #     :param lr_policy: name of learning rate policy: linear | step | plateau | cosine
    #     :param verbose: verbose for printing the networks, True or False
    #     :param kwargs: keyword arguments to set for learning rate
    #                    linear : epoch_count | n_epochs | n_epochs_decay
    #                    step   : n_decay_iters
    #                    plateau: 
    #                    cosine : n_epochs

    #     :return: None            
    #     """
    #     if len(schedulers) == 0:
    #         self.schedulers = [get_scheduler(optimizer, lr_policy, **kwargs) for optimizer in self.optimizers]
    #     else:
    #         self.schedulers = schedulers

    def set_mode(self, mode) -> None:
        """
        Make models eval mode during test time.
        
        :return: None
        """
        for name in self.model_names:
            if isinstance(name, str):
                net = getattr(self, name)
                
                if mode == 'train':
                    net.train()
                elif mode == 'eval':
                    net.eval()
                else:
                    raise Exception(f'Mode: {mode} is not recognized')

    def set_requires_grad(self, nets: list, requires_grad:bool = False):
        """
        Set requies_grad=Fasle for all the networks to avoid unnecessary computations
        
        :param nets (network list)  :   a list of networks
        :param requires_grad (bool) :   whether the networks require gradients or not
        """
        if not isinstance(nets, list):
            nets = [nets]
        for net in nets:
            if net is not None:
                for param in net.parameters():
                    param.requires_grad = requires_grad

    def test(self) -> None:
        """
        Forward function used in test time.
        
        This function wraps <forward> function in no_grad() so we don't save intermediate steps for backprop
        It also calls <compute_visuals> to produce additional visualization results

        :return: None
        """
        with torch.no_grad():
            self.forward()
            # self.compute_visuals()

    # def update_learning_rate(self) -> None:
    #     """
    #     Update learning rates for all the networks; called at the end of every epoch
    #     """

    #     old_lr = self.optimizers[0].param_groups[0]['lr']
    #     for scheduler in self.schedulers:
    #         if self.lr_policy == 'plateau':
    #             scheduler.step(self.metric) # TODO metric here is validation loss
    #         else:
    #             scheduler.step()

    #     lr = self.optimizers[0].param_groups[0]['lr']
    #     print('learning rate %.7f -> %.7f' % (old_lr, lr))

    def get_current_losses(self) -> OrderedDict:
        """
        Return traning losses / errors. train.py will print out these errors on console, and save them to a file
        """

        errors_ret = OrderedDict()
        for name in self.loss_names:
            if isinstance(name, str):
                errors_ret[name] = float(getattr(self, name))  # float(...) works for both scalar tensor and float number
        return errors_ret

    def print_networks(self, verbose=False) -> None:
        """
        Print the total number of parameters in the network and (if verbose) network architecture
        
        :param verbose (bool): if verbose: print the network architecture
        :return: None
        """

        print('----------------------------------------- Networks -----------------------------------------')
        for name in self.model_names:
            if isinstance(name, str):
                net = getattr(self, name)
                num_params = 0
                for param in net.parameters():
                    num_params += param.numel()
                print('-' * 10, f' [Network {name}] Total number of parameters : {format(num_params / 1e6, ".2f")} M ', '-' * 10)
                if verbose:
                    print(net)
        print('--------------------------------------------------------------------------------------------')


    # def get_current_checkpoint_name(self) -> str:
    #     '''
    #     Created current checkpoint name
    #     :return (str): checkpoint name without extension
    #     '''

    #     current_losses = self.get_current_losses()
    #     loss_str = '_'.join([str(i) + str(current_losses[i]) for i in current_losses])
    #     name = f'{self.__class__.__name__}_epoch{self.epoch}_{loss_str}'
    #     return name

    # def save_networks(self, dir_path:str) -> None:
    #     """
    #     Save all the networks to the disk.        
    #     :param dir (str): networks will be saved at given directory path
    #     :return: None
    #     """
    #     dir_path = Path(dir_path)
    #     dir_path = dir_path / self.get_current_checkpoint_name()
    #     dir_path.mkdir(parents=True, exist_ok=True)
    #     print(f"Saving Networks at {dir_path}")

    #     for name in self.model_names:
    #         if isinstance(name, str):
    #             file_name = f'net_{name}.pt'
    #             file_path = dir_path / file_name
    #             net = getattr(self, 'net_' + name)

    #             if 'cuda' in str(self.device):
    #                 torch.save(net.module.cpu(), file_path)
    #                 net.to(self.device)
    #             else:
    #                 torch.save(net, file_path)

    # def load_networks(self, dir_path: str) -> None:
    #     """
    #     Load all the networks from the disk.        
    #     :param dir_path: networks will be loaded from given directory path
    #     :return: None
    #     """
    #     dir_path = Path(dir_path)

    #     for name in self.model_names:
    #         if isinstance(name, str):
    #             file_name = f'net_{name}.pt'
    #             file_path = dir_path / file_name
    #             print(f'loading the model from {file_path}')        
    #             net = torch.load(file_path)
    #             net.to(self.device)
    #             if 'cuda' in str(self.device) :
    #                 net = torch.nn.DataParallel(net, self.gpu_ids)
    #             setattr(self, 'net_' + name, net)

    #     print("Networks have been loaded, make sure to compile again.")

    # def save_checkpoints(self, dir_path: str) -> None:
    #     '''
    #     Saves the checkpoint at given diretory path.
    #     :param dir_path (str): checkpoint will be saved at given file
    #     :return: None
    #     '''

    #     dir_path = Path(dir_path)
    #     dir_path = dir_path / self.get_current_checkpoint_name()
    #     dir_path.mkdir(parents=True, exist_ok=True)

    #     print(f'Saving checkpoint at {dir_path}')

    #     # For saving models
    #     for net_name in self.model_names:
    #         file_path = dir_path / ('net_' + net_name + '.pt')
    #         checkpoint = {}
    #         net = getattr(self, 'net_' + net_name)
            
    #         if isinstance(net, nn.DataParallel):
    #             state_dict = net.module.cpu().state_dict()
    #             net.to(self.device)
    #         else:
    #             state_dict = net.state_dict()

    #         checkpoint['net_' + net_name + '_state_dict'] = state_dict
    #         torch.save(checkpoint, file_path)

    #     # For saving Optimizers
    #     optims = {}
    #     for optim_name in self.optimizer_names:
    #         optim = getattr(self, 'optimizer_' + optim_name)
    #         optims['optimizer_' + optim_name + '_state_dict'] = optim.state_dict()
    #         torch.save(optims, dir_path / 'optimizers.pt')

    #     # Losses and Epoch will be saved in meta.pt
    #     # For saving Losses
    #     meta = {}
    #     current_losses = self.get_current_losses()
    #     for loss_name in current_losses:
    #         meta['loss_' + loss_name] = current_losses[loss_name]

    #     # For saving epoch
    #     meta['epoch'] = self.epoch

    #     # Sving the data
    #     torch.save(meta, dir_path / 'meta.pt')

    # def load_checkpoints(self, dir_path:str):
    #     '''
    #     Loads all the model checkpoints from the disk.

    #     :param dir_path (str): checkpoint will be loaded from given directory
    #     :return: None
    #     '''
    #     dir_path = Path(dir_path)
    #     print(f'Loading checkpoint from {dir_path}')

    #     # Loading Models
    #     for net_name in self.model_names:

    #         net = getattr(self, 'net_' + net_name)
    #         file_path = dir_path / ('net_' + net_name + '.pt')
    #         checkpoint = torch.load(file_path, map_location=str(self.device))
    #         state_dict = checkpoint['net_' + net_name + '_state_dict']
    #         if isinstance(net, nn.DataParallel):
    #             net = net.module            
    #         net.load_state_dict(state_dict)


    #     # Loading Optimizers
    #     optims = torch.load(dir_path / 'optimizers.pt')
    #     for optim_name in self.optimizer_names:

    #         optim = getattr(self, 'optimizer_' + optim_name)
    #         optim.load_state_dict(optims['optimizer_' + optim_name + '_state_dict'])

    #     # Loading Losses and Epoch from meta.pt
    #     meta = torch.load(dir_path / 'meta.pt')
    #     for loss_name in self.loss_names:
    #         setattr(self, 'loss_' + loss_name, meta['loss_' + loss_name])

    #     self.epoch = meta['epoch']

    # def compute_visuals(self):
    #     """
    #     Calculate additional output images for visdom and HTML visualization
    #     """
    #     pass

    # def get_image_paths(self):
    #     """ 
    #     Return image paths that are used to load current data
    #     """
    #     return self.image_paths

    # def get_current_visuals(self):
    #     """
    #     Return visualization images. train.py will display these images with visdom, and save the images to a HTML
    #     """
        
    #     visual_ret = OrderedDict()
    #     for name in self.visual_names:
    #         if isinstance(name, str):
    #             visual_ret[name] = getattr(self, name)
    #     return visual_ret



from typing import Optional, Union

import tensorflow as tf
import tensorflow_hub as hub

from deepdraken.utils import noise_sample

import numpy as np

class CompareGAN():

    '''
        Class for generating Compare GAN images.
    '''
    MODLES = {  'celeba-hq-1': 'https://tfhub.dev/google/compare_gan/model_1_celebahq128_resnet19/1',
                'celeba-hq-2': 'https://tfhub.dev/google/compare_gan/model_2_celebahq128_resnet19/1',
                'celeba-hq-3': 'https://tfhub.dev/google/compare_gan/model_5_celebahq128_resnet19/1',
                'celeba-hq-4': 'https://tfhub.dev/google/compare_gan/model_6_celebahq128_resnet19/1',
                'celeba-hq-5': 'https://tfhub.dev/google/compare_gan/model_9_celebahq128_resnet19/1',

                'lsun-bedroom-1':	'https://tfhub.dev/google/compare_gan/model_3_lsun_bedroom_resnet19/1',
                'lsun-bedroom-2':	'https://tfhub.dev/google/compare_gan/model_4_lsun_bedroom_resnet19/1',
                'lsun-bedroom-3': 'https://tfhub.dev/google/compare_gan/model_7_lsun_bedroom_resnet19/1',
                'lsun-bedroom-4': 'https://tfhub.dev/google/compare_gan/model_8_lsun_bedroom_resnet19/1',
                'lsun-bedroom-5': 'https://tfhub.dev/google/compare_gan/model_10_lsun_bedroom_resnet19/1',

                'cifar-10-1': 'https://tfhub.dev/google/compare_gan/model_11_cifar10_resnet_cifar/1',
                'cifar-10-2': 'https://tfhub.dev/google/compare_gan/model_12_cifar10_resnet_cifar/1',
                'cifar-10-3': 'https://tfhub.dev/google/compare_gan/model_13_cifar10_resnet_cifar/1',
                'cifar-10-4': 'https://tfhub.dev/google/compare_gan/model_14_cifar10_resnet_cifar/1',
                'cifar-10-5': 'https://tfhub.dev/google/compare_gan/model_15_cifar10_resnet_cifar/1'}

    def __init__(self, model_name: Optional[str] = 'cifar-10-1') -> None:

        self.model = hub.KerasLayer(self.MODLES[model_name], signature="generator")

    def __get_images(self, z: np.ndarray) -> np.ndarray:
        '''
        Method for generating images.
        :param z: noise vectors
        :return: generated images
        '''

        images = self.model({'z':z})
        images = np.asarray(images)
        images = np.clip(images * 255, 0, 255)
        images = np.uint8(images)
        return images

    def sample(self, num_samples:int) -> np.ndarray:
        '''
        Method for sampling images

        :param num_samples: number of images to generate
        :return: an array of generated images
        '''

        generated_samples = 0
        images = []


        while generated_samples < num_samples:
            # generating the noise
            z = noise_sample(64, 128) # compare gan expects the batch size to be 64
            # converting into array if needed
            z = np.asarray(z)

            image_batch = self.__get_images(z)
            images.append(image_batch)
            generated_samples += 64

        images = np.concatenate(images, axis=0)
        return images[:num_samples]